"""A collection of precanned workflows that can be achieved using the lower
level openchemistry API
"""

class Reaction(object):
    def __init__(self, equation):
        self._equation = equation

        (self._reactants, self._products) = equation.split('=>')
        self._reactants = [x.strip() for x in self._reactants.split('+')]
        self._products = [x.strip() for x in self._products.split('+')]

    @property
    def reactants(self):
        return self._reactants

    @property
    def products(self):
        return self._products

    @property
    def equation(self):
        return '%s => %s' % (' + '.join(self.reactants), ' + '.join(self.products))

def setup_reaction(equation):
    return Reaction(equation)

def compose_equation(equation, **vars):
    equation = Environment(loader=BaseLoader()).from_string(equation)

    return equation.render(**vars)



from jinja2 import Environment, BaseLoader
from girder_client import HttpError
import re
import avogadro

from ._girder import GirderClient
from ._molecule import Molecule
from ._calculation import GirderMolecule, CalculationResult, AttributeInterceptor, _fetch_calculation
from ._data import CjsonProvider, AvogadroProvider
from ._utils import fetch_or_create_queue

_inchi_key_regex = re.compile("^([0-9A-Z\-]+)$")

def _is_inchi_key(identifier):
    return len(identifier) == 27 and identifier[25] == '-' and \
        _inchi_key_regex.match(identifier)

def _find_molecule(identifier=None, inchi=None, smiles=None):
    if inchi:
        return _find_molecule_by_inchi(inchi)

    if smiles:
        return _find_molecule_by_smiles(smiles)

    # InChiKey?
    if _is_inchi_key(identifier):
        return _find_molecule_by_inchikey(identifier)

    # Finally, if identifier is something else (a name), try cactus
    if identifier:
        return _find_molecule_using_cactus(identifier)

    return None

def _find_molecule_by_inchi(inchi):
    params = {
        'inchi': inchi
    }
    return _find_molecule_using_girder(params)

def _find_molecule_by_smiles(smiles):
    params = {
        'smiles': smiles
    }
    return _find_molecule_using_girder(params)

def _find_molecule_by_inchikey(inchikey):
    try:
        return GirderClient().get('molecules/inchikey/%s' % inchikey)
    except HttpError as ex:
        if ex.status == 404:
            return _find_molecule_using_cactus(inchikey)
        else:
            raise

def _find_molecule_using_cactus(identifier):
    params = {
        'cactus': identifier
    }
    molecules = GirderClient().get('molecules/search', parameters=params)
    # Just pick the first
    if len(molecules) > 0:
        return molecules[0]
    else:
        return None

def _find_molecule_using_girder(params):
    molecules = GirderClient().get('molecules', parameters=params)
    if len(molecules) > 0:
        return molecules[0]
    else:
        return None

def _calculation_monitor(taskflow_ids):
    try:
        from ._notebook import CalculationMonitor
        table = CalculationMonitor({
            'taskFlowIds': taskflow_ids,
            'girderToken': GirderClient().token,
            'girderApiUrl': GirderClient().api_url
        })
    except ImportError:
        # Outside notebook just print message
        table = 'Pending calculations .... '

    return table

def import_structure(smiles=None, inchi=None, cjson=None):
    # If the smiles begins with 'InChI=', then it is actually an inchi instead
    if smiles and smiles.startswith('InChI='):
        inchi = smiles
        smiles = None

    params = {}
    if smiles:
        params['smiles'] = smiles
    elif inchi:
        params['inchi'] = inchi
    elif cjson:
        params['cjson'] = cjson
    else:
        raise Exception('SMILES, InChI, or CJson must be provided')

    molecule = GirderClient().post('molecules', json=params)

    if not molecule:
        raise Exception('Molecule could not be imported with params', params)

    return GirderMolecule(molecule['_id'], molecule['cjson'])

def find_molecule(identifier=None, inchi=None, smiles=None):
    molecule = _find_molecule(identifier, inchi, smiles)
    if molecule is None:
        raise Exception('Unable to find a molecule with the provided identifiers.')
    return GirderMolecule(molecule['_id'], molecule.get('cjson'))

def find_calculation(molecule, image_name=None, input_parameters=None, input_geometry=None):
    calculation = _fetch_calculation(molecule._id, image_name, input_parameters, input_geometry)
    if calculation is None:
        raise Exception('Unable to find a matching calculation in the database')
    return CalculationResult(calculation['_id'])

def find_structure(identifier=None, image_name=None, input_parameters=None, input_geometry=None, inchi=None, smiles=None):
    molecule = find_molecule(identifier, inchi, smiles)

    # If we have been provided basis, theory or functional it means the user is
    # looking for a calculation, otherwise they're looking for a molecule
    is_calc_query = (image_name is not None and input_parameters is not None)

    if is_calc_query:
        return find_calculation(molecule, image_name, input_parameters, input_geometry)
    else:
        return molecule

def load(data):
    if isinstance(data, dict):
        provider = CjsonProvider(data)
    elif isinstance(data, avogadro.core.Molecule):
        provider = AvogadroProvider(data)
    else:
        raise TypeError("Load accepts either a cjson dict, or an avogadro.core.Molecule")
    return Molecule(provider)

def monitor(results):
    taskflow_ids = []

    for result in results:
        if hasattr(result, '_properties'):
            props = result._properties
            if isinstance(props, AttributeInterceptor):
                props = props.unwrap()
            if isinstance(props, dict):
                taskflow_id = props.get('taskFlowId')
                if taskflow_id is not None:
                    taskflow_ids.append(taskflow_id)

    return _calculation_monitor(taskflow_ids)

def queue():
    if GirderClient().host is None:
        import warnings
        warnings.warn("Cannot displaying pending calculations, the notebook is not running in a Girder environment")
        return

    queue = fetch_or_create_queue(GirderClient())
    running = []
    pending = []

    for taskflow_id, status in queue['taskflows'].items():
        if status == 'pending':
            pending.append(taskflow_id)
        elif status == 'running':
            running.append(taskflow_id)

    taskflow_ids = running + pending

    return _calculation_monitor(taskflow_ids)

def find_spectra(identifier, stype='IR', source='NIST'):
    """Find spectra in source database

    Parameters
    ----------
    identifier : str
        Inchi string.
    stype : str
        Type of spectrum to query.
    source : str
        Database to query. Supported are: 'NIST'
    """
    source = source.lower()

    params = {
            'inchi' : identifier,
            'spectrum_type' : stype,
            'source' : source
    }

    spectra = GirderClient().get('experiments', parameters=params)

    return spectra



def _cclib_to_cjson_basis(basis):
    shell_type_map = {
        's': 0,
        'p': 1,
        'd': 2,
        'f': 3,
        'g': 4,
        'h': 5,
        'i': 6,
        'k': 7,
        'l': 8,
        'm': 9,
        'n': 10,
        'o': 11
    }
    coefficients = []
    exponents = []
    primitives_per_shell = []
    shell_to_atom_map = []
    shell_types = []
    for i_atom, atom_basis in enumerate(basis):
        for shell in atom_basis:
            l_label, primitives = shell
            n_primitives = len(primitives)
            shell_to_atom_map.append(i_atom)
            primitives_per_shell.append(n_primitives)
            shell_types.append(shell_type_map[l_label.lower()])
            for primitive in primitives:
                exponents.append(primitive[0])
                coefficients.append(primitive[1])
    cjson_basis = {
        'coefficients': coefficients,
        'exponents': exponents,
        'primitivesPerShell': primitives_per_shell,
        'shellToAtomMap': shell_to_atom_map,
        'shellTypes': shell_types
    }
    return cjson_basis

def _cclib_to_cjson_mocoeffs(coeffs):
    cjson_coeffs = []
    # only take the orbitals at the end of the optimization
    for mo in coeffs[-1]:
        cjson_coeffs.extend(mo)
    return cjson_coeffs

def _cclib_to_cjson_vibdisps(vibdisps):
    cjson_vibdisps = []
    for vibdisp in vibdisps:
        cjson_vibdisps.append(list(vibdisp.flatten()))
    return cjson_vibdisps



import openchemistry as oc

calc_setup = {
    'basis': '3-21g', # What fields in db?
    'theory': 'dft'
}

mymol = oc.find_structure('RYYVLZVUVIJVGH-UHFFFAOYSA-N')
result = mymol.optimize(**calc_setup)
result.structure.show()

result = mymol.frequencies(**calc_setup)
result.frequencies.show()
freq_table = result.frequencies.table()
print(freq_table)



import os
import requests
import re
import json
import hashlib

import avogadro

from jsonpath_rw import parse
from IPython.lib import kernel

def fetch_or_create_queue(girder_client):
    params = {'name': 'oc_queue'}
    queue = girder_client.get('queues', parameters=params)

    if (len(queue) > 0):
        queue = queue[0]
    else:
        params = {'name': 'oc_queue', 'maxRunning': 5}
        queue = girder_client.post('queues', parameters=params)

    return queue

def lookup_file(girder_client, jupyterhub_url):
    """
    Utility function to lookup the Girder file that the current notebook is running
    from.
    """
    connection_file_path = kernel.get_connection_file()
    me = girder_client.get('user/me')
    login = me['login']
    kernel_id = re.search('kernel-(.*).json', connection_file_path).group(1)

    # Authenticate with jupyterhub so we can get the cookie
    r = requests.post('%s/hub/login' % jupyterhub_url, data={'Girder-Token': girder_client.token}, allow_redirects=False)
    r.raise_for_status()
    cookies = r.cookies

    url = "%s/user/%s/api/sessions" % (jupyterhub_url, login)
    r = requests.get(url, cookies=cookies)
    r.raise_for_status()


    sessions = r.json()
    matching = [s for s in sessions if s['kernel']['id'] == kernel_id]
    path = matching[0]['path']
    name = os.path.basename(path)

    # Logout
    url = "%s/hub/logout" % jupyterhub_url
    r = requests.get(url, cookies=cookies)


    return girder_client.resourceLookup('user/%s/Private/oc/notebooks/%s/%s' % (login, path, name))

def calculate_mo(cjson, mo):
    if isinstance(mo, str):
        mo = mo.lower()
        if mo.lower() in ['homo', 'lumo']:
            # Electron count might be saved in several places...
            path_expressions = [
                'orbitals.electronCount',
                'basisSet.electronCount',
                'properties.electronCount'
            ]
            matches = []
            for expr in path_expressions:
                matches.extend(parse(expr).find(cjson))
            if len(matches) > 0:
                electron_count = matches[0].value
            else:
                raise Exception('Unable to access electronCount')

            # The index of the first orbital is 0, so homo needs to be
            # electron_count // 2 - 1
            if mo.lower() == 'homo':
                mo = int(electron_count / 2) - 1
            elif mo.lower() == 'lumo':
                mo = int(electron_count / 2)
        else:
            raise ValueError('Unsupported mo: %s' % mo)

    mol = avogadro.core.Molecule()
    conv = avogadro.io.FileFormatManager()
    conv.read_string(mol, json.dumps(cjson), 'cjson')
    # Do some scaling of our spacing based on the size of the molecule.
    atom_count = mol.atom_count()
    spacing = 0.30
    if atom_count > 50:
        spacing = 0.5
    elif atom_count > 30:
        spacing = 0.4
    elif atom_count > 10:
        spacing = 0.33
    cube = mol.add_cube()
    # Hard wiring spacing/padding for now, this could be exposed in future too.
    cube.set_limits(mol, spacing, 4)
    gaussian = avogadro.core.GaussianSetTools(mol)
    gaussian.calculate_molecular_orbital(cube, mo)

    return json.loads(conv.write_string(mol, "cjson"))['cube']

def hash_object(obj):
    return hashlib.sha512(json.dumps(obj, sort_keys=True).encode()).hexdigest()

def camel_to_space(s):
    s = re.sub(r"""
        (            # start the group
            # alternative 1
        (?<=[a-z])  # current position is preceded by a lower char
                    # (positive lookbehind: does not consume any char)
        [A-Z]       # an upper char
                    #
        |   # or
            # alternative 2
        (?<!\A)     # current position is not at the beginning of the string
                    # (negative lookbehind: does not consume any char)
        [A-Z]       # an upper char
        (?=[a-z])   # matches if next char is a lower char
                    # lookahead assertion: does not consume any char
        )           # end the group""",
    r' \1', s, flags=re.VERBOSE)
    return s[0:1].upper() + s[1:]

def parse_image_name(image_name):
    split = image_name.split(":")
    if len(split) > 2:
        raise ValueError('Invalid Docker image name provided')
    elif len(split) == 1:
        repository = split[0]
        tag = 'latest'
    else:
        repository, tag = split

    return repository, tag



import os
from girder_client import GirderClient as GC

from ._utils import lookup_file

class GirderClient(object):

    class __GirderClient(object):
        def __init__(self):
            girder_host = os.environ.get('GIRDER_HOST')
            girder_port = os.environ.get('GIRDER_PORT')
            girder_scheme = os.environ.get('GIRDER_SCHEME', 'http')
            girder_api_root = os.environ.get('GIRDER_API_ROOT', '/api/v1')
            girder_api_key = os.environ.get('GIRDER_API_KEY')
            girder_token = os.environ.get('GIRDER_TOKEN')
            # First check if we have a public url set
            girder_api_url = os.environ.get('GIRDER_PUBLIC_API_URL')
            if girder_api_url is None:
                girder_api_url = '%s://%s%s/%s' % (
                    girder_scheme, girder_host, ':%s' % girder_port if girder_port else '',
                    girder_api_root )
            app_base_url = os.environ.get('APP_BASE_URL')
            cluster_id = os.environ.get('CLUSTER_ID')
            jupyterhub_url = os.environ.get('OC_JUPYTERHUB_URL')

            self.file = None
            self.client = None
            self.host = None
            self.api_url = girder_api_url
            self.app_url = app_base_url
            self.cluster_id = cluster_id

            if girder_host:
                self.client = GC(host=girder_host, port=girder_port,
                                            scheme=girder_scheme, apiRoot=girder_api_root)
                self.host = girder_host

                if girder_api_key is not None:
                    self.client.authenticate(apiKey=girder_api_key)
                elif girder_token is not None:
                    self.client.token = girder_token

                if jupyterhub_url is not None:
                    self.file = lookup_file(self.client, jupyterhub_url)

        def __getattr__(self, name):
            return getattr(self.client, name)

    instance = None

    def __init__(self):
        if GirderClient.instance is None:
            GirderClient.instance = GirderClient.__GirderClient()

    def __getattr__(self, name):
        return getattr(GirderClient.instance, name)



from abc import ABC, abstractmethod
import urllib.parse

from ._utils import hash_object, camel_to_space

class Visualization(ABC):
    def __init__(self, provider):
        self._provider = provider
        self._params = {}

    @abstractmethod
    def show(self, viewer='moljs', spectrum=False, volume=False, isosurface=False, menu=True, mo=None, iso=None, transfer_function=None, mode=-1, play=False, alt=None):
        self._params = {
            'moleculeRenderer': viewer,
            'showSpectrum': spectrum,
            'showVolume': volume,
            'showIsoSurface': isosurface,
            'showMenu': menu,
            'mo': mo,
            'isoValue': iso,
            'mode': mode,
            'play': play,
            **self._transfer_function_to_params(transfer_function)
        }
        try:
            from ._notebook import CJSON
            return CJSON(self._provider.cjson, **self._params)
        except ImportError:
            # Outside notebook print CJSON
            if alt is None:
                print(self._provider.cjson)
            else:
                print(alt)

    @abstractmethod
    def data(self):
        return self._provider.cjson

    def url(self):
        url = self._provider.url
        if url is None:
            raise ValueError('The current structure is not coming from a calculation')

        if self._params:
            url = '%s?%s' % (url, urllib.parse.urlencode(self._params))
        try:
            from IPython.display import Markdown
            return Markdown('[%s](%s)' % (url, url))
        except ImportError:
            # Outside notebook just print the url
            print(url)

    def _transfer_function_to_params(self, transfer_function):
        '''
        transfer_function = {
            "colormap": {
                "colors": [[r0, g0, b0], [r1, g1, b1], ...],
                "points": [p0, p1, ...]
            },
            "opacitymap": {
                "opacities": [alpha0, alpha1, ...],
                "points": [p0, p1, ...]
            }
        }
        '''
        params = {}

        if transfer_function is None or not isinstance(transfer_function, dict):
            return params

        colormap = transfer_function.get('colormap')
        if colormap is not None:
            colors = colormap.get('colors')
            points = colormap.get('points')
            if colors is not None:
                params['colors'] = colors
                params['activeMapName'] = 'Custom'
            if points is not None:
                params['colorsX'] = points

        opacitymap = transfer_function.get('opacitymap')
        if opacitymap is not None:
            opacities = opacitymap.get('opacities')
            points = opacitymap.get('points')
            if opacities is not None:
                params['opacities'] = opacities
            if points is not None:
                params['opacitiesX'] = points

        return params

class Structure(Visualization):

    def show(self, viewer='moljs', menu=True, **kwargs):
        return super(Structure, self).show(viewer=viewer, menu=menu, **kwargs)

    def data(self):
        return self._provider.cjson

class Vibrations(Visualization):

    def show(self, viewer='moljs', spectrum=True, menu=True, mode=-1, play=True, **kwargs):
        return super(Vibrations, self).show(viewer=viewer, spectrum=spectrum, menu=menu, mode=mode, play=play, **kwargs)

    def data(self):
        return self._provider.vibrations

    def table(self):
        vibrations = self._provider.vibrations
        try:
            from IPython.display import Markdown
            table = self._md_table(vibrations)
            return Markdown(table)
        except ImportError:
            # Outside notebook print CJSON
            print(vibrations)

    def _md_table(self, vibrations):
        import math
        table = '''### Normal Modes
| # | Frequency | Intensity |
|------|-------|-------|'''
        frequencies = vibrations.get('frequencies', [])
        intensities = vibrations.get('intensities', [])

        n = len(frequencies)
        if len(intensities) != n:
            intensities = None

        for i, freq in enumerate(frequencies):
            try:
                freq = float(freq)
            except ValueError:
                freq = math.nan

            intensity = math.nan if intensities is None else intensities[i]
            try:
                intensity = float(intensity)
            except ValueError:
                intensity = math.nan

            table += '\n| %s | %.2f | %.2f |' % (
                i,
                freq,
                intensity
            )

        return table

class Orbitals(Visualization):

    def show(self, viewer='moljs', volume=False, isosurface=True, menu=True, mo='homo', iso=0.05, transfer_function=None, **kwargs):
        self._provider.load_orbital(mo)
        return super(Orbitals, self).show(viewer=viewer, volume=volume, isosurface=isosurface, menu=menu, mo=mo, iso=iso, transfer_function=transfer_function)

    def data(self):
        whitelist = ['orbitals', 'properties']
        output = {}
        for item in whitelist:
            output[item] = self._provider.cjson[item]

        return output

class Properties(Visualization):

    def show(self, **kwargs):
        cjson = self._provider.cjson
        properties = cjson.get('properties', {})
        try:
            from IPython.display import Markdown
            table = self._md_table(properties)
            return Markdown(table)
        except ImportError:
            # Outside notebook print CJSON
            print(properties)

    def data(self):
        return self._provider.cjson.get('properties', {})

    def _md_table(self, properties):
        import math
        table = '''### Calculated Properties
| Name | Value |
|------|-------|'''

        for prop, value in properties.items():
            try:
                value = float(value)
            except ValueError:
                value = math.nan
            table += '\n| %s | %.2f |' % (
                camel_to_space(prop),
                value
            )

        return table



from ._utils import hash_object, camel_to_space, parse_image_name
from .io import Psi4Reader, NWChemJsonReader, CjsonReader
from .api import load, find_structure, find_calculation, find_molecule, monitor, queue, find_spectra



import os
import inspect
from jsonpath_rw import parse

from ._girder import GirderClient
from ._molecule import Molecule
from ._data import MoleculeProvider, CalculationProvider
from ._utils import fetch_or_create_queue, hash_object, parse_image_name

class GirderMolecule(Molecule):
    '''
    Derived version that allows calculations to be initiated on using Girder
    '''
    def __init__(self, _id, cjson=None):
        if cjson is None:
            cjson = GirderClient().get('molecules/%s/cjson' % _id)
        super(GirderMolecule, self).__init__(MoleculeProvider(cjson, _id))
        self._id = _id

    def calculate(self, image_name, input_parameters, input_geometry=None, run_parameters=None, force=False):
        molecule_id = self._id
        calculation = _fetch_or_submit_calculation(molecule_id, image_name, input_parameters, input_geometry, run_parameters, force)
        pending = parse('properties.pending').find(calculation)
        if pending:
            pending = pending[0].value

        taskflow_id = parse('properties.taskFlowId').find(calculation)
        if taskflow_id:
            taskflow_id = taskflow_id[0].value
        else:
            taskflow_id = None

        calculation = CalculationResult(calculation['_id'], calculation['properties'], molecule_id)

        if pending:
            calculation = PendingCalculationResultWrapper(calculation, taskflow_id)

        return calculation

    def energy(self, image_name, input_parameters, input_geometry=None, run_parameters=None, force=False):
        params = {'task': 'energy'}
        params.update(input_parameters)
        return self.calculate(image_name, params, input_geometry, run_parameters, force)

    def optimize(self, image_name, input_parameters, input_geometry=None, run_parameters=None, force=False):
        params = {'task': 'optimize'}
        params.update(input_parameters)
        return self.calculate(image_name, params, input_geometry, run_parameters, force)

    def frequencies(self, image_name, input_parameters, input_geometry=None, run_parameters=None, force=False):
        params = {'task': 'frequencies'}
        params.update(input_parameters)
        return self.calculate(image_name, params, input_geometry, run_parameters, force)

    def set_name(self, name):
        body = {
            'name': name
        }
        GirderClient().patch('molecules/%s' % self._id, json=body)

class CalculationResult(Molecule):

    def __init__(self, _id=None, properties=None, molecule_id=None):
        super(CalculationResult, self).__init__(CalculationProvider(_id, molecule_id))
        self._id = _id
        self._properties = properties
        self._molecule_id = molecule_id

    def data(self):
        return self._provider.cjson

    @property
    def frequencies(self):
        import warnings
        warnings.warn("Use the 'vibrations' property to display normal modes")
        return self.vibrations

    def delete(self):
        return _delete_calculation(self._id)

class AttributeInterceptor(object):
    def __init__(self, wrapped, value, intercept_func=lambda : True):
        self._wrapped = wrapped
        self._value = value
        self._intercept_func = intercept_func

    def unwrap(self):
        return self._wrapped

    def __getattr__(self, name):
        # Use object's implementation to get attributes, otherwise
        # we will get recursion
        _wrapped = object.__getattribute__(self, '_wrapped')
        _value = object.__getattribute__(self, '_value')
        _intercept_func = object.__getattribute__(self, '_intercept_func')

        attr = object.__getattribute__(_wrapped, name)
        if _intercept_func():
            if inspect.ismethod(attr):
                def pending(*args, **kwargs):
                    return _value
                return pending
            else:
                return AttributeInterceptor(attr, _value, _intercept_func)
        else:
            return attr

class PendingCalculationResultWrapper(AttributeInterceptor):
    def __init__(self, calculation, taskflow_id=None):
        try:
            from ._notebook import CalculationMonitor
            if taskflow_id is None:
                taskflow_id = calculation._properties['taskFlowId']

            table = CalculationMonitor({
                'taskFlowIds': [taskflow_id],
                'girderToken': GirderClient().token,
                'girderApiUrl': GirderClient().api_url
            })
        except ImportError:
            # Outside notebook just print message
            table = 'Pending calculations .... '

        # Only intercept when the taskflow is not complete
        def intercept():
            return _fetch_taskflow_status(taskflow_id) != 'complete'

        super(PendingCalculationResultWrapper, self).__init__(calculation,
                                                              table, intercept)

def _fetch_calculation(molecule_id, image_name, input_parameters, input_geometry=None):
    repository, tag = parse_image_name(image_name)
    parameters = {
        'moleculeId': molecule_id,
        'inputParametersHash': hash_object(input_parameters),
        'imageName': '%s:%s' % (repository, tag)
    }

    if input_geometry:
        parameters['inputGeometryHash'] = hash_object(input_geometry)

    calculations = GirderClient().get('calculations', parameters)

    if len(calculations) < 1:
        return None

    return calculations[0]

def _nersc():
    return os.environ.get('OC_SITE') == 'NERSC'

def _submit_calculation(cluster_id, pending_calculation_id, image_name, run_parameters):
    if cluster_id is None and not _nersc():
        # Try to get demo cluster
        params = {
            'type': 'trad'
        }
        clusters = GirderClient().get('clusters', params)

        if len(clusters) > 0:
            cluster_id = clusters[0]['_id']
        else:
            raise Exception('Unable to submit calculation, no cluster configured.')

    if run_parameters is None:
        run_parameters = {}

    repository, tag = parse_image_name(image_name)

    # Create the taskflow
    queue = fetch_or_create_queue(GirderClient())

    body = {
        'taskFlowClass': 'taskflows.OpenChemistryTaskFlow',
        'meta': {
            'calculationId': pending_calculation_id,
            'image': {
                'repository': repository,
                'tag': tag
            }
        }
    }

    taskflow = GirderClient().post('taskflows', json=body)

    # Start the taskflow
    body = {
        'input': {
            'calculation': {
                '_id': pending_calculation_id
            }
        },
        'image': {
            'repository': repository,
            'tag': tag
        },
        'runParameters': run_parameters
    }

    if cluster_id is not None:
        body['cluster'] = {
            '_id': cluster_id
        }
    elif _nersc():
        body['cluster'] = {
            'name': 'cori'
        }

    GirderClient().put('queues/%s/add/%s' % (queue['_id'], taskflow['_id']), json=body)
    GirderClient().put('queues/%s/pop' % queue['_id'], parameters={'multi': True})

    return taskflow['_id']

def _fetch_taskflow_status(taskflow_id):
    r = GirderClient().get('taskflows/%s/status' % taskflow_id)

    return r['status']

def _create_pending_calculation(molecule_id, image_name, input_parameters, input_geometry=None):
    repository, tag = parse_image_name(image_name)

    notebooks = []
    if GirderClient().file is not None:
        notebooks.append(GirderClient().file['_id'])

    body = {
        'moleculeId': molecule_id,
        'cjson': None,
        'public': True,
        'properties': {
            'pending': True
        },
        'input': {
            'parameters': input_parameters,
        },
        'image': {
            'repository': repository,
            'tag': tag
        },
        'notebooks': notebooks
    }

    if input_geometry is not None:
        body['input']['geometry'] = input_geometry

    calculation = GirderClient().post('calculations', json=body)

    return calculation

def _delete_calculation(calculation_id):
    GirderClient().delete('calculations/%s' % calculation_id)

def _fetch_or_submit_calculation(molecule_id, image_name, input_parameters, input_geometry=None, run_parameters=None, force=False):
    calculation = _fetch_calculation(molecule_id, image_name, input_parameters, input_geometry)
    taskflow_id = None

    if calculation is None or force:
        calculation = _create_pending_calculation(molecule_id, image_name, input_parameters, input_geometry)
        taskflow_id = _submit_calculation(GirderClient().cluster_id, calculation['_id'], image_name, run_parameters)
        # Patch calculation to include taskflow id
        props = calculation['properties']
        props['taskFlowId'] = taskflow_id
        calculation = GirderClient().put('calculations/%s/properties' % calculation['_id'], json=props)
    else:
        # If we already have a calculation tag it with this notebooks id
        notebooks = calculation.setdefault('notebooks', [])
        if GirderClient().file is not None:
            notebooks.append(GirderClient().file['_id'])

        body = {
            'notebooks': notebooks
        }
        GirderClient().patch('calculations/%s/notebooks' % calculation['_id'],
                            json=body)

    return calculation
